from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from pprint import pprint

import subprocess

import ROOT

from nittygriddy import utils


def merge(args):
    # check if we are in a output dirname
    if not ((os.path.isfile("./ConfigureTrain.C") or
             os.path.isfile("./MLTrainDefinition.cfg")) and
            os.path.isfile("./GetSetting.C") and
            os.path.isfile("./run.C")):
        raise ValueError("This command needs to be run from an output directory")

    if not utils.check_alien_token():
        print("No valid alien token present. Run `alien-token-init` before nittygriddy")
        quit()

    ROOT.gROOT.LoadMacro(r'GetSetting.C')
    if ROOT.GetSetting("runmode") != "grid":
        raise ValueError("The data in this folder was not run on the grid!")

    # root does not like it if the stdout is piped and it uses some shell functionality
    # thus, using call on the whole string and shell True
    if args.mergemode == 'online':
        cmd = r'root -b -l -q -x "run.C(\"merge_online\")"'
        subprocess.check_call(cmd, shell=True)
    elif args.mergemode == 'offline':
        cmd = r'root -b -l -q -x "run.C(\"merge_offline\")"'
        subprocess.check_call(cmd, shell=True)

    if args.mergemode == 'download':
        workdir = os.path.dirname(os.path.abspath("./run.C")).split("/")[-1]
        alien_workdir = os.path.join(utils.find_user_grid_dir(), workdir)
        files = utils.find_latest_merge_results(alien_workdir)
        print("The following files will be downloaded:")
        pprint(files)
        for alien_path in files:
            local_path = alien_path.split(workdir)[-1].lstrip("/")
            try:
                utils.download_file(alien_path, local_path)
            except ValueError:
                print("{} exists and is not redownloaded.".format(local_path))
            except OSError as e:
                # If there was an error in the download, just print, but keep going
                print(e.message)

    if args.mergemode == 'clean':
        workdir = os.path.dirname(os.path.abspath("./run.C")).split("/")[-1]
        alien_workdir = os.path.join(utils.find_user_grid_dir(), workdir)
        files = utils.find_sources_of_merged_files(utils.find_latest_merge_results(alien_workdir))
        folders = [os.path.split(f)[0] for f in files]
        if not folders:
            print("No files to be cleaned up")
            return
        print("The following folders will be __DELETED__:")
        pprint(folders)
        shall_delete = utils.yn_choice("Do you really want to delete these Folders? This may take a while")
        if shall_delete:
            # its not possible to pass all arguments at once (too many)
            # Hence, pass 10 at a time
            for slice_start in range(len(folders))[::10]:
                cmd = ['alien_rmdir'] + folders[slice_start:(slice_start + 10)]
                # alien_rmdir always seems to return a non-zero exit code... Yeah!
                subprocess.call(cmd)
        # see if the deletion was successful:
        left_files = utils.find_sources_of_merged_files(utils.find_latest_merge_results(alien_workdir))
        if left_files:
            print("The following files could not be deleted:")
            pprint(left_files)

    if args.mergemode == 'unmerge':
        workdir = os.path.dirname(os.path.abspath("./run.C")).split("/")[-1]
        alien_workdir = os.path.join(utils.find_user_grid_dir(), workdir)
        subprocess.call(['alien_rm', os.path.join(alien_workdir, 'output/*/Stage_*.xml')])
        subprocess.call(['alien_rm', os.path.join(alien_workdir, 'output/*/AnalysisResults.root')])
        subprocess.call(['alien_rm', os.path.join(alien_workdir, 'output/*/root_archive.zip')])
        subprocess.call(['alien_rmdir', os.path.join(alien_workdir, 'output/*/Stage_*')])


def create_subparsers(subparsers):
    # Create merge sub-parser
    description_merge = """Merge related tasks connected to a previously run grid analysis. Must be
    executed from output folder of that analysis."""
    parser_merge = subparsers.add_parser('merge', description=description_merge)
    parser_merge.add_argument('mergemode',
                              choices=('online', 'offline', 'download', 'clean', 'unmerge'),
                              help=("Merge files online, offline or download the latest merge results. "
                                    "Use `clean` to delete everything except the last merge stage. "
                                    "Use this if you only wnat to keep the final result to safe space. "
                                    "`unmerge` deletes all merge stages but kepes "
                                    "the raw data. Use this if you want to re-merge your "
                                    "data after more jobs finished. Don't use this in combination "
                                    "with `clean` or you will be sorry."))
    parser_merge.set_defaults(op=merge)



"""
Create the command line parser
"""

import argparse
from nittygriddy import run, merge, datasets, new, profile


def create_parser():
    """
    Creates the parser object

    By using a function, the parser is also easily available for unittests
    """
    class formatter_class(argparse.ArgumentDefaultsHelpFormatter,
                          argparse.RawTextHelpFormatter):
        pass

    parser = argparse.ArgumentParser(formatter_class=formatter_class)
    parser.add_argument(
        '-v', '--verbose', action='store_true', default=False)
    subparsers = parser.add_subparsers()
    run.create_subparsers(subparsers)
    merge.create_subparsers(subparsers)
    datasets.create_subparser(subparsers)
    new.create_subparsers(subparsers)
    profile.create_subparsers(subparsers)
    return parser



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tempfile
import os
from os import path
import subprocess
from . import utils


def profile(args):
    """
    Run the poor-mans-profiler and open a browser to see the produced svg
    """
    pid = str(args.pid)
    tmpdir = tempfile.gettempdir()
    # make sure we don't have old files floating around
    gdb_file = path.join(tmpdir, pid + ".gdb")
    svg_file = path.join(tmpdir, pid + ".svg")
    for f in [gdb_file, svg_file]:
        try:
            os.remove(f)
        except:
            pass
    nitty_root = utils._internal_files_dir()
    poor_mans = path.join(nitty_root, "poormans.sh")
    print("See the profiling result at {}".format(svg_file))
    p = subprocess.Popen(['bash', poor_mans, str(args.nsamples), pid, gdb_file, svg_file, nitty_root])
    try:
        p.wait()
    except KeyboardInterrupt:
        p.kill()


def create_subparsers(subparsers):
    description = """Profile a running nitty-process. The profiler produces a svg file which can be view in the browsers. It updates every few seconds, so be patient. This feature depends on gdb. If you get "ERROR: No stack counts found" messages, you might not have set the proper permissions for gdb. Google for the file /etc/sysctl.d/10-ptrace.conf"""
    parser_new = subparsers.add_parser('profile', description=description)
    parser_new.add_argument('pid', help='PID of the process you want to profile')
    parser_new.add_argument('--nsamples', help='Number of samples to be taken', default=50)
    parser_new.set_defaults(op=profile)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import shutil

from unittest import TestCase, skip
import subprocess

from nittygriddy import utils, settings, parser
from nittygriddy.alienTokenError import AlienTokenError


class Test_downloader(TestCase):
    """
    These tests might be broken for downlaods through TFile::Copy. I
    think it has something to do with the environment variable and how
    they are set when running through nosetests
    """
    def setUp(self):
        test_local_dir = "~/lhc_data_test/"
        settings["local_data_dir"] = test_local_dir
        try:
            # make sure that we really only delete a __test__ dir
            if 'test' in test_local_dir:
                os.rmdir(test_local_dir)
        except OSError:
            pass

    def test_invalid_dataset(self):
        self.assertRaises(KeyError, utils.download_dataset, "invalid_dataset", 5)

    @skip("Skip download test")
    def test_download_one_file(self):
        try:
            shutil.rmtree("/tmp/nitty_test")
        except:
            pass
        remote_path = "/alice/cern.ch/user/c/cbourjau/nitty_test/AnalysisResults.root"
        utils.download_file(remote_path,
                            "/tmp/nitty_test/AnalysisResults.root")
        self.assertTrue(os.path.isfile("/tmp/nitty_test/AnalysisResults.root"))
        shutil.rmtree("/tmp/nitty_test")
        # without specifiying dest file name:
        utils.download_file(remote_path,
                            "/tmp/nitty_test/")
        self.assertTrue(os.path.isfile("/tmp/nitty_test/AnalysisResults.root"))
        shutil.rmtree("/tmp/nitty_test")

    # @skip("Skip download test")
    def test_downloaded_something(self):
        utils.check_alien_token()
        utils.download_dataset("LHC10h_AOD160", 0.001)

    @skip("Skip download test")
    def test_download_files_from_archive(self):
        try:
            shutil.rmtree("/tmp/nitty_test")
        except:
            pass
        utils.download_from_grid_archive(
            '/alice/sim/2012/LHC12a11a/137686/AOD157/0001/root_archive.zip',
            '/tmp/nitty_test/alice/sim/2012/LHC12a11a/137686/AOD157/0001/root_archive.zip',
        )
        self.assertTrue(os.path.isfile(
            '/tmp/nitty_test/alice/sim/2012/LHC12a11a/137686/AOD157/0001/AliAOD.root'))
        shutil.rmtree("/tmp/nitty_test")


class Test_grid_features(TestCase):
    def test_latest_aliphysics(self):
        self.assertNotEqual(utils.get_latest_aliphysics(), "")

    def test_par_file(self):
        self.assertRaises(ValueError, utils.prepare_par_files, "I_dont_exists.par", "test_folder")

    def test_find_user_grid_dir(self):
        grid_home = utils.find_user_grid_dir()
        self.assertRegexpMatches(grid_home, '/alice/cern.ch/user/\w/\w+/')

    def find_latest_merge_results(self):
        # a test dir set up under alien:///alice/cern.ch/user/c/cbourjau/
        workdir = "nitty_test"
        # try with not absolute workdir:
        self.assertRaises(ValueError, utils.find_latest_merge_results, workdir)
        # with absolut alien path
        alien_workdir = os.path.join(utils.find_user_grid_dir(), workdir)
        self.assertEqual(utils.find_latest_merge_results(alien_workdir),
                         ['/alice/cern.ch/user/c/cbourjau/nitty_test/AnalysisResults.root'])

    def test_find_merge_sources(self):
        workdir = "nitty_test"
        alien_workdir = os.path.join(utils.find_user_grid_dir(), workdir)
        merge_results = utils.find_latest_merge_results(alien_workdir)
        self.assertEqual(utils.find_sources_of_merged_files(merge_results),
                         ['/alice/cern.ch/user/c/cbourjau/nitty_test/001/AnalysisResults.root',
                          '/alice/cern.ch/user/c/cbourjau/nitty_test/002/AnalysisResults.root'])

    def test_find_associated_archive_files(self):
        urls = utils.find_associated_archive_files(
            datadir="/alice/sim/2012/LHC12a11a",
            run_number_prefix="",
            runs=[137686, 138534, 138653, 139038, 139437],
            data_pattern="AOD157/*/AliAOD.root")
        self.assertGreater(len(urls), 0)


class Test_environment(TestCase):
    """
    These tests require a valid alien token.
    """
    def test_alien_token_valid(self):
        self.assertTrue(utils.check_alien_token())

    @skip("Skip token invalid test")
    def test_alien_token_invalid(self):
        cmd = ['alien-token-destroy']
        subprocess.check_output(cmd)
        self.assertRaises(AlienTokenError, utils.check_alien_token)


class Test_create_GetSetting_c_file(TestCase):
    """
    Test the create of the GetSetting.C file
    """
    def test_create_getSetting_c(self):
        import tempfile
        tmp_dir = tempfile.gettempdir()
        _parser = parser.create_parser()
        args = _parser.parse_args(['run', 'lite', 'LHC12a11a'])
        utils.prepare_get_setting_c_file(tmp_dir, args)


class Test_invalid_user_dataset_def(TestCase):
    """
    Test a bunch of stuff concerning the user created datasets
    """
    def missing_fields(self):
        # Empty one should pass
        ds = {}
        utils.validate_dataset(ds)
        # Inner entry is not a dictionary
        ds = {'blub': ''}
        self.assertRaises(ValueError, utils.validate_dataset, ds)
        # Inner entry does not have the keys but is a dict
        ds = {'blub': dict()}
        self.assertRaises(ValueError, utils.validate_dataset, ds)


class Test_cli(TestCase):
    def test_datasets(self):
        p = parser.create_parser()
        args = p.parse_args(['datasets', '-l'])
        args.op(args)
